import event_model
from ._version import get_versions
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from pymongo import UpdateOne
import pymongo
from threading import Event
import time
import queue
import bson

__version__ = get_versions()['version']
del get_versions


class Serializer(event_model.DocumentRouter):
    """
    Insert bluesky documents into MongoDB using an embedded data model.

    This embedded data model has three collections: header, event,
    datum. The header collection includes start, stop, descriptor,
    and resource documents. The event_pages are stored in the event colleciton,
    and datum_pages are stored in the datum collection.

    This Serializer ensures that when the stop document or close request is
    received all documents we be written to the database.

    This Serializer assumes that documents have been previously validated
    according to the bluesky event-model.

    Note that this Seralizer does not share the standard Serializer
    name or signature common to suitcase packages because it can only write
    via pymongo, not to an arbitrary user-provided buffer.

    Examples
    --------
    >>> from bluesky import RunEngine
    >>> from bluesky.plans import scan
    >>> from mongobox import MongoBox
    >>> from ophyd.sim import det, motor
    >>> from suitcase.mongo_embedded import Serializer

    >>> # Create a sandboxed mongo instance.
    >>> mongo_box = MongoBox()
    >>> mongo_box.start()

    >>> # Get a reference to the mongo database.
    >>> db = mongo_box.client().db

    >>> # Create the Serializer
    >>> serializer = Serializer(db)

    >>> RE = RunEngine({})
    >>> RE.subscribe(serializer)

    >>> # Generate example data.
    >>> RE(scan([det], motor, 1, 10, 10))
    """

    def __init__(self, db, num_threads=1, queue_size=100,
                 embedder_size=1000000, page_size=5000000,
                 max_insert_time=5, **kwargs):

        """
        Insert documents into MongoDB using an embedded data model.

        Parameters
        ----------
        db: pymongo database
        num_theads: int, optional
            number of workers that read from the buffer and write to the
            database. Must be 5 or less. Default is 1.
        queue_size: int, optional
            maximum size of the queue.
        page_size: int, optional
            the document size for event_page and datum_page documents. The
            maximum event/datum_page size is embedder_size + page_size.
            embedder_size + page_size must be less than 15000000.The
            default is 5000000.
        embedder_size: int, optional
            maximum size of the embedder
        max_insert_time: int, optional
            maximum time that the workers will wait before doing a database
            insert.
        """

        # There is no performace improvment for more than 10 threads. Tests
        # validate for upto 10 threads.
        if num_threads > 10 or num_threads < 1:
            raise ValueError(f"num_threads must be between 1 and 10"
                             "inclusive.")

        if page_size < 1000:
            raise ValueError(f"page_size must be >= 1000")

        # Maximum size of a document in mongo is 16MB. buffer_size + page_size
        # defines the biggest document that can be created.
        if embedder_size + page_size > 15000000:
            raise ValueError(f"embedder_size: {embedder_size} + page_size:"
                             "{page_size} is greater then 15000000.")

        self._QUEUE_SIZE = queue_size
        self._EMBED_SIZE = embedder_size
        self._PAGE_SIZE = page_size
        self._MAX_INSERT = max_insert_time
        self._QUEUE_TIMEOUT = 0.2
        self._db = db
        self._event_queue = queue.Queue(maxsize=self._QUEUE_SIZE)
        self._datum_queue = queue.Queue(maxsize=self._QUEUE_SIZE)
        self._event_embedder = Embedder('event', self._EMBED_SIZE)
        self._datum_embedder = Embedder('datum', self._EMBED_SIZE)
        self._kwargs = kwargs
        self._start_found = False
        self._run_uid = None
        self._frozen = False
        self._count = Event()
        self._worker_error = None
        self._stop_doc = None

        # _event_count and _datum_count are used for setting first/last_index
        # fields of event and datum pages
        self._event_count = defaultdict(lambda: 0)
        self._datum_count = defaultdict(lambda: 0)

        # These counters are used to track the total number of events and datum
        # that have been successfully inserted into the database.
        self._db_event_count = defaultdict(lambda: 0)
        self._db_datum_count = defaultdict(lambda: 0)

        # Start workers.
        self._event_executor = ThreadPoolExecutor(max_workers=1)
        self._datum_executor = ThreadPoolExecutor(max_workers=1)
        self._count_executor = ThreadPoolExecutor(max_workers=1)
        self._event_executor.submit(self._event_worker)
        self._datum_executor.submit(self._datum_worker)
        self._count_executor.submit(self._count_worker)

        self._create_indexes()

    def _create_indexes(self):
        """
        Create indexes on the various collections.
         If the index already exists, this has no effect.
        """
        self._db.header.create_index('resources.uid', sparse=True)
        self._db.header.create_index('resources.resource_id')  # legacy
        self._db.header.create_index(
            [('start.uid', pymongo.DESCENDING)], unique=True, sparse=True)
        self._db.header.create_index(
            [('start.time', pymongo.DESCENDING),
             ('start.scan_id', pymongo.DESCENDING)],
            unique=False, background=True)
        self._db.header.create_index([("$**", "text")])
        self._db.header.create_index('stop.run_start', unique=True, sparse=True)
        self._db.header.create_index('stop.uid', unique=True, sparse=True)
        self._db.header.create_index(
            [('stop.time', pymongo.DESCENDING)], unique=False,
            background=True, sparse=True)
        self._db.header.create_index(
            [('descriptors.uid', pymongo.DESCENDING)], unique=True, sparse=True)
        self._db.header.create_index(
            [('descriptors.run_start', pymongo.DESCENDING),
             ('time', pymongo.DESCENDING)],
            unique=False, background=True)
        self._db.header.create_index(
            [('descriptors.time', pymongo.DESCENDING)],
            unique=False, background=True)
        self._db.event.create_index(
            [('uid', pymongo.DESCENDING)], unique=True, sparse=True)
        self._db.event.create_index(
            [('descriptor', pymongo.DESCENDING),
             ('time.0', pymongo.ASCENDING)],
            unique=False, background=True)
        self._db.datum.create_index('datum_id', unique=True, sparse=True)
        self._db.datum.create_index('resource')

    def __call__(self, name, doc):
        # Before inserting into mongo, convert any numpy objects into built-in
        # Python types compatible with pymongo.
        sanitized_doc = event_model.sanitize_doc(doc)
        if self._worker_error:
            raise RuntimeError("Worker exception: ") from self._worker_error
        if self._frozen:
            raise RuntimeError("Cannot insert documents into "
                               "frozen Serializer.")

        return super().__call__(name, sanitized_doc)

    def _try_wrapper(f):
        from functools import wraps
        @wraps(f)
        def inner(self):
            try:
                f(self)
            except Exception as error:
                self._worker_error = error
                raise
        return inner

    @_try_wrapper
    def _event_worker(self):
        # Gets events from the queue, embedds them, and writes them to the
        # database.
        last_push = 0
        event = None

        # When a stop document is received 'False' is pushed on to the
        # queue, this signals the worker to finish.
        while event is not False:
            last_push = time.monotonic()
            do_push = False
            try:
                if event is None:
                    event = self._event_queue.get(timeout=self._QUEUE_TIMEOUT)
            except queue.Empty:
                do_push = True
            else:
                # embedder.insert() returns None if the document is inserted,
                # and returns the document, if embedder is full.
                if event is not False:
                    event = self._event_embedder.insert(event)
            if (
                    event is not None
                    or event is False
                    or do_push
                    or time.monotonic() > (last_push + self._MAX_INSERT)):
                if not self._event_embedder.empty():
                    event_dump, dump_sizes = self._event_embedder.dump()
                    self._bulkwrite_event(event_dump, dump_sizes)
                    for descriptor, event_page in event_dump.items():
                        self._db_event_count['count_' + descriptor] += len(
                                event_page['seq_num'])
                last_push = time.monotonic()
                do_push = False

    @_try_wrapper
    def _datum_worker(self):
        # Gets datum from the queue, embedds them, and writes them to the
        # database.

        last_push = 0
        datum = None

        # When a stop document is received 'False' is pushed on to the
        # queue, this signals the worker to finish.
        while datum is not False:
            last_push = time.monotonic()
            do_push = False
            try:
                if datum is None:
                    datum = self._datum_queue.get(timeout=self._QUEUE_TIMEOUT)
            except queue.Empty:
                do_push = True
            else:
                # embedder.insert() returns None if the document is inserted,
                # and returns the document, if embedder is full.
                if datum is not False:
                    datum = self._datum_embedder.insert(datum)
            if (
                    datum is not None
                    or datum is False
                    or do_push
                    or time.monotonic() > (last_push + self._MAX_INSERT)):

                if not self._datum_embedder.empty():
                    datum_dump, dump_sizes = self._datum_embedder.dump()
                    self._bulkwrite_datum(datum_dump, dump_sizes)
                    for resource, datum_page in datum_dump.items():
                        self._db_datum_count['count_' + resource] += len(
                                datum_page['datum_id'])
                last_push = time.monotonic()
                do_push = False

    @_try_wrapper
    def _count_worker(self):
        # Updates event_count and datum_count in the header document

        last_event_count = defaultdict(lambda: 0)
        last_datum_count = defaultdict(lambda: 0)

        while not self._frozen:
            self._count.wait(timeout=5)
            # Only updates the header if the count has changed.
            if (
                    (sum(self._db_event_count.values()) >
                     sum(last_event_count.values()))
                    or (sum(self._db_datum_count.values()) >
                        sum(last_datum_count.values()))
               ):
                self._db.header.update_one(
                    {'run_id': self._run_uid},
                    {'$set': {**dict(self._db_event_count),
                              **dict(self._db_datum_count)}})

                last_event_count = self._db_event_count
                last_datum_count = self._db_datum_count

    def start(self, doc):
        self._check_start(doc)
        self._run_uid = doc['uid']
        self._insert_header('start', doc)
        self._insert_header('event_count', doc)
        self._insert_header('datum_count', doc)
        return doc

    def stop(self, doc):
        self._stop_doc = doc
        self.close()
        return doc

    def descriptor(self, doc):
        self._insert_header('descriptors', doc)
        return doc

    def resource(self, doc):
        self._insert_header('resources', doc)
        return doc

    def event(self, doc):
        self._event_queue.put(doc)
        return doc

    def datum(self, doc):
        self._datum_queue.put(doc)
        return doc

    def event_page(self, doc):
        doc_size = len(bson.BSON.encode(doc))
        self._bulkwrite_event({doc['descriptor']: doc},
                              {doc['descriptor']: doc_size})
        return doc

    def datum_page(self, doc):
        doc_size = len(bson.BSON.encode(doc))
        self._bulkwrite_datum({doc['resource']: doc},
                              {doc['resource']: doc_size})
        return doc

    def close(self):
        self.finalize(self._run_uid)

    def finalize(self, run_uid):
        """
        Finalize insertion of the run.
        """
        # Freeze the serializer.
        self._frozen = True
        self._event_queue.put(False)
        self._datum_queue.put(False)

        # Interupt the count worker sleep
        self._count.set()

        self._count_executor.shutdown(wait=True)
        self._event_executor.shutdown(wait=True)
        self._datum_executor.shutdown(wait=True)

        self._set_header('event_count', sum(self._event_count.values()))
        self._set_header('datum_count', sum(self._datum_count.values()))

        if self._worker_error:
            raise RuntimeError("Worker exception: ") from self._worker_error

        # Raise exception if buffers are not empty.
        assert self._event_queue.empty()
        assert self._datum_queue.empty()
        assert self._event_embedder.empty()
        assert self._datum_embedder.empty()

        # Insert the stop doc.
        self._insert_header('stop', self._stop_doc)

    def _insert_header(self, name,  doc):
        """
        Inserts header document into the run's header document.
        """
        self._db.header.update_one({'run_id': self._run_uid},
                                   {'$push': {name: doc}},
                                   upsert=True)

    def _set_header(self, name,  doc):
        """
        Inserts header document into the run's header document.
        """
        self._db.header.update_one({'run_id': self._run_uid},
                                   {'$set': {name: doc}})

    def _bulkwrite_datum(self, datum_buffer, dump_sizes):
        """
        Bulk writes datum_pages to Mongo datum collection.
        """
        operations = [self._updateone_datumpage(resource, datum_page,
                                                dump_sizes[resource])
                      for resource, datum_page in datum_buffer.items()]
        self._db.datum.bulk_write(operations, ordered=False)

    def _bulkwrite_event(self, event_buffer, dump_sizes):
        """
        Bulk writes event_pages to Mongo event collection.
        """
        operations = [self._updateone_eventpage(descriptor, event_page,
                                                dump_sizes[descriptor])
                      for descriptor, event_page in event_buffer.items()]
        self._db.event.bulk_write(operations, ordered=False)

    def _updateone_eventpage(self, descriptor_id, event_page, size):
        """
        Creates the UpdateOne command that gets used with bulk_write.
        """
        event_size = size

        data_string = {'data.' + key: {'$each': value_array}
                       for key, value_array in event_page['data'].items()}

        timestamp_string = {'timestamps.' + key: {'$each': value_array}
                            for key, value_array
                            in event_page['timestamps'].items()}

        filled_string = {'filled.' + key: {'$each': value_array}
                         for key, value_array in event_page['filled'].items()}

        update_string = {**data_string, **timestamp_string, **filled_string}

        count = len(event_page['seq_num'])
        self._event_count[descriptor_id] += count

        return UpdateOne(
            {'descriptor': descriptor_id, 'size': {'$lt': self._PAGE_SIZE}},
            {'$push': {'uid': {'$each': event_page['uid']},
                       'time': {'$each': event_page['time']},
                       'seq_num': {'$each': event_page['seq_num']},
                       **update_string},
             '$inc': {'size': event_size},
             '$min': {'first_index': self._event_count[descriptor_id] - count},
             '$max': {'last_index': self._event_count[descriptor_id] - 1}},
            upsert=True)

    def _updateone_datumpage(self, resource_id, datum_page, size):
        """
        Creates the UpdateOne command that gets used with bulk_write.
        """
        datum_size = size

        kwargs_string = {'datum_kwargs.' + key: {'$each': value_array}
                         for key, value_array
                         in datum_page['datum_kwargs'].items()}

        count = len(datum_page['datum_id'])
        self._datum_count[resource_id] += count

        return UpdateOne(
            {'resource': resource_id, 'size': {'$lt': self._PAGE_SIZE}},
            {'$push': {'datum_id': {'$each': datum_page['datum_id']},
                       **kwargs_string},
             '$inc': {'size': datum_size},
             '$min': {'first_index': self._datum_count[resource_id] - count},
             '$max': {'last_index': self._datum_count[resource_id] - 1}},
            upsert=True)

    def _check_start(self, doc):
        if self._start_found:
            raise RuntimeError(
                "The serializer in suitcase-mongo expects "
                "documents from one run only. Two `start` documents were "
                "received.")
        else:
            self._start_found = True


class Embedder():

    """
    Embedder embeds normalized bluesky documents.

    "embedding" refers to combining multiple documents from a stream of
    documents into a single document, where the values of matching keys are
    stored as a list, or dictionary of lists. "embedding" converts event docs
    to event_pages, or datum doc to datum_pages. event_pages and datum_pages
    are defined by the bluesky event-model.

    Events with different descriptors, or datum with different resources are
    stored in separate embedded documents. Embedder uses a defaultdict so new
    embedded documents are automatically created when they are needed. The dump
    method returns the embedded dictionary. This mechanism manages the lifetime
    of the event streams in the buffer.

    The doc_type argument which can be either 'event' or 'datum'.
    The the details of the embedding differ for event and datum documents.

    Internally the embedder is a dictionary that maps event decriptors to
    event_pages or datum resources to datum_pages.

    Parameters
    ----------
    doc_type: str
        {'event', 'datum'}
    max_size: int
        Maximum embedder size in bytes.

    Attributes
    ----------
    current_size: int
        Current size of the embedded documents.

    """

    def __init__(self, doc_type, max_size):
        self._embedder = defaultdict(lambda: defaultdict(lambda:
                                                         defaultdict(list)))
        self.current_size = 0
        self.stream_size = defaultdict(lambda: 0)

        if (max_size >= 1000) and (max_size <= 15000000):
            self._max_size = max_size
        else:
            raise ValueError(f"Invalid max_size {max_size}, "
                             "max_size must be between 1000 and "
                             "15000000 inclusive.")

        # Event docs and datum docs are embedded differently, this configures
        # the buffer for the specified document type.
        if doc_type == "event":
            self._array_keys = set(["seq_num", "time", "uid"])
            self._dataframe_keys = set(["data", "timestamps", "filled"])
            self._stream_id_key = "descriptor"
        elif doc_type == "datum":
            self._array_keys = set(["datum_id"])
            self._dataframe_keys = set(["datum_kwargs"])
            self._stream_id_key = "resource"
        else:
            raise ValueError(f"Invalid doc_type {doc_type}, doc_type must "
                             "be either 'event' or 'datum'")

    def dump(self):
        """
        Get everything that has been embedded  and clear the buffer.

        Returns
        -------
        embedder_dump: dict
            A dictionary that maps event descriptor to event_page, or a
            dictionary that maps datum resource to datum_page.
        """
        # Get a reference to the current dict, create a new dict.
        embedder_dump = self._embedder
        self._embedder = defaultdict(lambda: defaultdict(lambda:
                                                         defaultdict(list)))
        dump_sizes = dict(self.stream_size)
        self.stream_size = defaultdict(lambda: 0)
        self.current_size = 0
        return embedder_dump, dump_sizes

    def insert(self, doc):
        """
        Embeds a bluesky event or datum document.
        Parameters
        ----------
        doc: json
            A validated bluesky event or datum document.
        Returns
        -------
        result: bool
            True if insert is successful, False if it failed.
        """
        doc_size = len(bson.BSON.encode(doc))
        if doc_size > self._max_size:
            raise ValueError(f"Document size is too large to fit in the "
                             f"embedder. doc_size={doc_size}, "
                             f"embedder_size={self._max_size}, "
                             f"doc_uid={doc['uid']}")

        if (self.current_size + doc_size) > self._max_size:
            return doc

        for key, value in doc.items():
            if key in self._array_keys:
                self._embedder[doc[self._stream_id_key]][key] = list(
                    self._embedder[doc[self._stream_id_key]][key])
                self._embedder[doc[self._stream_id_key]][key].append(value)
            elif key in self._dataframe_keys:
                for inner_key, inner_value in doc[key].items():
                    (self._embedder[doc[self._stream_id_key]][key]
                        [inner_key].append(inner_value))
            else:
                self._embedder[doc[self._stream_id_key]][key] = value

        self.current_size += doc_size
        self.stream_size[doc[self._stream_id_key]] += doc_size
        return None

    def empty(self):
        return not self.current_size



# Tests should generate (and then clean up) any files they need for testing. No
# binary files should be included in the repository.
import json
import event_model
from suitcase.mongo_embedded import Serializer
import pytest


def test_export(db_factory, example_data):
    """
    Test suitcase-mongo-embedded serializer with default parameters.
    """
    permanent_db = db_factory()
    serializer = Serializer(permanent_db)
    run(example_data, serializer, permanent_db)
    if not serializer._frozen:
        serializer.close()


def test_multithread(db_factory, example_data):
    """
    Test suitcase-mongo-embedded serializer with multiple worker threads.
    """
    permanent_db = db_factory()
    serializer = Serializer(permanent_db, num_threads=5)
    run(example_data, serializer, permanent_db)
    if not serializer._frozen:
        serializer.close()


def test_smallbuffer(db_factory, example_data):
    """
    Test suitcase-mongo-embedded serializer with a small buffer.
    """
    permanent_db = db_factory()
    serializer = Serializer(permanent_db, embedder_size=3000)
    run(example_data, serializer, permanent_db)
    if not serializer._frozen:
        serializer.close()


def test_smallqueue(db_factory, example_data):
    """
    Test suitcase-mongo-embedded serializer with a small buffer.
    """
    permanent_db = db_factory()
    serializer = Serializer(permanent_db, queue_size=1)
    run(example_data, serializer, permanent_db)
    if not serializer._frozen:
        serializer.close()


def test_smallpage(db_factory, example_data):
    """
    Test suitcase-mongo-embedded serializer with a small mongo page saize.
    """
    permanent_db = db_factory()
    serializer = Serializer(permanent_db, page_size=10000)
    run(example_data, serializer, permanent_db)
    if not serializer._frozen:
        serializer.close()


def test_evil_db(db_factory, example_data):
    """
    Test suitcase-mongo-embedded serializer with a db that raises an exception
    on bulk_write.
    """
    def evil_func(*args, **kwargs):
        raise RuntimeError

    permanent_db = db_factory()
    serializer = Serializer(permanent_db)
    serializer._bulkwrite_event = evil_func
    serializer._bulkwrite_datum = evil_func
    with pytest.raises(RuntimeError):
        run(example_data, serializer, permanent_db)
    if not serializer._frozen:
        serializer.close()


def run(example_data, serializer, permanent_db):
    """
    Testbench for suitcase-mongo-embedded serializer.
    This stores all documents that are going to the serializer into a
    dictionary. After the run completes, it then queries the permanent
    mongo database, and reads the documents to a separate dictionary. The two
    dictionaries are checked to see if they match.
    """
    run_dict = {'start': {}, 'stop': {}, 'descriptor': [],
                'resource': [], 'event': [], 'datum': []}
    documents = example_data()
    mongo_serializer = serializer
    for item in documents:

        # Fix formatting for JSON.
        item = event_model.sanitize_doc(item)
        # Send the bluesky doc to the serializer
        mongo_serializer(*item)

        # Bulk_event/datum need to be converted to a list of events/datum
        # before inserting in the run_dict.
        if item[0] in {'bulk_events', 'bulk_datum'}:
            pages = bulk_to_pages(*item)
            doc_list = pages_to_list(pages)
            for doc in doc_list:
                run_dict[doc[0]].append(doc[1])
        else:
            if item[0] in {'event_page', 'datum_page'}:
                doc_list = page_to_list(*item)
                for doc in doc_list:
                    run_dict[doc[0]].append(doc[1])
            else:
                if type(run_dict.get(item[0])) == list:
                    run_dict[item[0]].append(item[1])
                else:
                    run_dict[item[0]] = item[1]

    # Read the run from the mongo database and store in a dict.
    frozen_run_dict = run_list_to_dict(get_embedded_run(
                                permanent_db, run_dict['start']['uid']))

    # Sort the event field of each dictionary. With multiple streams, the
    # documents that don't go through the serializer don't appear to be sorted
    # correctly.
    if len(run_dict['event']):
        run_dict['event'] = sorted(run_dict['event'],
                                   key=lambda x: x['descriptor'])
        frozen_run_dict['event'] = sorted(frozen_run_dict['event'],
                                          key=lambda x: x['descriptor'])
    # Compare the two dictionaries.
    assert (json.loads(json.dumps(run_dict, sort_keys=True))
            == json.loads(json.dumps(frozen_run_dict, sort_keys=True)))


def run_list_to_dict(embedded_run_list):
    """
    Converts a run from the mongo database to a dictionary.
    """
    run_dict = {'start': {},
                'stop': {},
                'descriptor': [],
                'resource': [],
                'event': [],
                'datum': []}

    header = embedded_run_list[0][1]
    run_dict['start'] = header['start'][0]
    run_dict['stop'] = header['stop'][0]
    run_dict['descriptor'] = header.get('descriptors', [])
    run_dict['resource'] = header.get('resources', [])

    for name, doc in embedded_run_list[1:]:
        if name == 'event':
            run_dict['event'] += list(event_model.unpack_event_page(doc))
        elif name == 'datum':
            run_dict['datum'] += list(event_model.unpack_datum_page(doc))

    return run_dict


def get_embedded_run(db, run_uid):
    """
    Gets a run from a database. Returns a list of the run's documents.
    """
    run = list()

    # Get the header.
    header = db.header.find_one({'run_id': run_uid}, {'_id': False})
    if header is None:
        raise RuntimeError(f"Run not found {run_uid}")

    run.append(('header', header))

    # Get the events.
    if 'descriptors' in header.keys():
        for descriptor in header['descriptors']:
            run += [('event', doc) for doc in
                    db.event.find({'descriptor': descriptor['uid']},
                                  {'_id': False})]

    # Get the datum.
    if 'resources' in header.keys():
        for resource in header['resources']:
            run += [('datum', doc) for doc in
                    db.datum.find({'resource': resource['uid']},
                                  {'_id': False})]
    return run


def bulk_to_pages(name, doc):
    """
    Converts bulk_events/datum to event/datum_page.
    """
    key_map = {'bulk_events': 'event_page', 'bulk_datum': 'datum_page'}

    if name == 'bulk_events':
        doc = event_model.bulk_events_to_event_pages(doc)
    elif name == 'bulk_datum':
        doc = event_model.bulk_datum_to_datum_pages(doc)

    page_list = [[key_map[name], item] for item in doc]

    return page_list


def pages_to_list(pages):
    """
    Converts event/datum_page to event/datum lists.
    """
    doc_list = []

    for page in pages:
        if page[0] == 'event_page':
            doc_list.extend([['event', event] for event
                            in event_model.unpack_event_page(page[1])])
        if page[0] == 'datum_page':
            doc_list.extend([['datum', datum] for datum
                            in event_model.unpack_datum_page(page[1])])

    return doc_list


def page_to_list(name, page):
    """
    Converts event/datum_page to event/datum lists.
    """
    doc_list = []

    if name == 'event_page':
        doc_list.extend([['event', event] for event
                        in event_model.unpack_event_page(page)])
    if name == 'datum_page':
        doc_list.extend([['datum', datum] for datum
                        in event_model.unpack_datum_page(page)])

    return doc_list



# Tests should generate (and then clean up) any files they need for testing. No
# binary files should be included in the repository.


from suitcase.mongo_normalized import Serializer


def test_export(db_factory, example_data):
    documents = example_data()
    metadatastore_db = db_factory()
    asset_registry_db = db_factory()
    serializer = Serializer(metadatastore_db, asset_registry_db)
    for item in documents:
        serializer(*item)



# This separate fixtures module allows external libraries (e.g.
# intake-bluesky-mongo) to import and reuse this fixtures without importing
# *all* the fixtures used in conftest and the dependencies that they carry.
import pymongo
import pytest
import uuid


@pytest.fixture()
def db_factory(request):
    def inner():
        database_name = f'test-{str(uuid.uuid4())}'
        uri = f'mongodb://localhost:27017/'
        client = pymongo.MongoClient(uri)

        def drop():
            client.drop_database(database_name)

        request.addfinalizer(drop)
        return client[database_name]
    return inner



import event_model
import pymongo
from ._version import get_versions


__version__ = get_versions()['version']
del get_versions


class Serializer(event_model.DocumentRouter):
    def __init__(self, metadatastore_db, asset_registry_db):
        """
        Insert documents into MongoDB using layout v1.

        This layout uses a separate Mongo collection per document type and a
        separate Mongo document for each logical document.

        Note that this Seralizer does not share the standard Serializer
        name or signature common to suitcase packages because it can only write
        via pymongo, not to an arbitrary user-provided buffer.

        Parameters
        ----------
        metadatastore_db : pymongo.Database or URI
        asset_registry_db : pymongo.Database or URI
        """
        if isinstance(metadatastore_db, str):
            mds_db = _get_database(metadatastore_db)
        else:
            mds_db = metadatastore_db
        if isinstance(asset_registry_db, str):
            assets_db = _get_database(asset_registry_db)
        else:
            assets_db = asset_registry_db
        self._run_start_collection = mds_db.get_collection('run_start')
        self._run_stop_collection = mds_db.get_collection('run_stop')
        self._event_descriptor_collection = mds_db.get_collection(
                                                        'event_descriptor')
        self._event_collection = mds_db.get_collection('event')

        self._resource_collection = assets_db.get_collection('resource')
        self._datum_collection = assets_db.get_collection('datum')

        self._metadatastore_db = mds_db
        self._asset_registry_db = assets_db

        self._create_indexes()

    def _create_indexes(self):
        """
        Create indexes on the various collections.

        If the index already exists, this has no effect.
        """
        self._resource_collection.create_index('uid')
        self._resource_collection.create_index('resource_id')  # legacy
        # TODO: Migrate all Resources to have a RunStart UID, and then make a
        # unique index on:
        # [('uid', pymongo.ASCENDING), ('run_start', pymongo.ASCENDING)]
        self._datum_collection.create_index('datum_id', unique=True)
        self._datum_collection.create_index('resource')
        self._run_start_collection.create_index(
            [('uid', pymongo.DESCENDING)], unique=True)
        self._run_start_collection.create_index(
            [('time', pymongo.DESCENDING), ('scan_id', pymongo.DESCENDING)],
            unique=False, background=True)
        self._run_start_collection.create_index([("$**", "text")])
        self._run_stop_collection.create_index('run_start', unique=True)
        self._run_stop_collection.create_index('uid', unique=True)
        self._run_stop_collection.create_index(
            [('time', pymongo.DESCENDING)], unique=False, background=True)
        self._run_stop_collection.create_index([("$**", "text")])
        self._event_descriptor_collection.create_index(
            [('uid', pymongo.DESCENDING)], unique=True)
        self._event_descriptor_collection.create_index(
            [('run_start', pymongo.DESCENDING), ('time', pymongo.DESCENDING)],
            unique=False, background=True)
        self._event_descriptor_collection.create_index(
            [('time', pymongo.DESCENDING)], unique=False, background=True)
        self._event_descriptor_collection.create_index([("$**", "text")])
        self._event_collection.create_index(
            [('uid', pymongo.DESCENDING)], unique=True)
        self._event_collection.create_index(
            [('descriptor', pymongo.DESCENDING), ('time', pymongo.ASCENDING)],
            unique=False, background=True)

    def __call__(self, name, doc):
        # Before inserting into mongo, convert any numpy objects into built-in
        # Python types compatible with pymongo.
        sanitized_doc = event_model.sanitize_doc(doc)
        return super().__call__(name, sanitized_doc)

    def start(self, doc):
        self._run_start_collection.insert_one(doc)

    def descriptor(self, doc):
        self._event_descriptor_collection.insert_one(doc)

    def resource(self, doc):
        self._resource_collection.insert_one(doc)

    def event(self, doc):
        self._event_collection.insert_one(doc)

    def event_page(self, doc):
        # Unpack an EventPage into Events and do the actual insert inside
        # the `event` method. (This is the oppose what DocumentRouter does by
        # default.)

        event_method = self.event  # Avoid attribute lookup in hot loop.
        filled_events = []

        for event_doc in event_model.unpack_event_page(doc):
            filled_events.append(event_method(event_doc))

    def datum(self, doc):
        self._datum_collection.insert_one(doc)

    def datum_page(self, doc):
        # Unpack an DatumPage into Datum and do the actual insert inside
        # the `datum` method. (This is the oppose what DocumentRouter does by
        # default.)

        datum_method = self.datum  # Avoid attribute lookup in hot loop.
        filled_datums = []

        for datum_doc in event_model.unpack_datum_page(doc):
            filled_datums.append(datum_method(datum_doc))

    def stop(self, doc):
        self._run_stop_collection.insert_one(doc)

    def __repr__(self):
        # Display connection info in eval-able repr.
        return (f'{type(self).__name__}('
                f'metadatastore_db={self._metadatastore_db!r}, '
                f'asset_registry_db={self._asset_registry_db!r})')


def _get_database(uri):
    client = pymongo.MongoClient(uri)
    try:
        # Called with no args, get_database() returns the database
        # specified in the client's uri --- or raises if there was none.
        # There is no public method for checking this in advance, so we
        # just catch the error.
        return client.get_database()
    except pymongo.errors.ConfigurationError as err:
        raise ValueError(
            f"Invalid client: {client} "
            f"Did you forget to include a database?") from err
